# HoleHealing

# The script opens the shape-file and looks through in him all polgyons with a holes. 
# Then the radius of the hole's polygon and its area is calculated.
# If the radius of a hole's polygon is less than specified value the hole is removed. 
# In case search of a polygon's radius finishes unsuccessfully 
# as criterion of a hole removal its area is used

# Constants are set by user
DataDir <- 'c:/Users/Const/Documents/Brazil/Data/'
VariantDir <- 'ARB_2016_R/'

# Shape file with input data
InFileName  <- paste0(DataDir, VariantDir, 'ARB_2016_RoadCut_R9.shp')

# Output Shape file 
OutFileName  <- paste0(DataDir, VariantDir, 'ARB_2016_Dehole_R9.shp')

# Shape file with problem holes where hole's polygon for  
# which it wasn't succeeded to calculate radius will be placed
BadHoleFileName  <- paste0(DataDir, VariantDir, 'ARB_2016_BadHole_R9.shp')

# Shape file with problem fragments where fragment's multipolygon for 
# which it wasn't succeeded to calculate radius will be placed
BadFragmentFileName  <- paste0(DataDir, VariantDir, 'ARB_2016_BadFragment_R9.shp')

# The minimum radius of a hole (m), is less which they are deleted
MaxHoleRadius <- 200

# The minimum hole's area (hectare), is less which they are deleted. 
# It is applied only in case of failure of the hole's radius search algorithm
MaxHoleArea <- 10
#--------------------------------------------------------------------------------------------
getLargestPolyPart <- function(multipol) 
{
  pols <- multipol@polygons[[1]]@Polygons
  if (length(pols) > 1) 
  {
    areas <- lapply(pols, function(x) x@area)
    pols <- pols[which.max(areas)]
    lstPolygons <- list(Polygons(pols, ID = "1"))
    CRC <- CRS(as.character(proj4string(multipol)))
    multipol <- SpatialPolygons(lstPolygons, proj4string = CRC)
  }
  return(multipol)
}
#--------------------------------------------------------------------------------------------
labpos.buffer <- function(pol) 
{
  init <- 0
  estep <- sqrt(gArea(pol)/pi)/10
  repeat 
  {
    repeat 
    {
      r <- init+estep
      pol.b <- gBuffer(pol, width = -r)
      if (is.null(pol.b)) 
        estep <- estep/2
      else 
      {
        A <- gArea(pol.b)
        if (A <= 0) 
          estep <- estep/2
        else 
          break
      }
    }
    polb <- getLargestPolyPart(pol.b)
    if (gContains(pol, gConvexHull(polb))) 
      break
    else 
      init <- init+estep
  }
  return(coordinates(pol.b))
}
#--------------------------------------------------------------------------------------------
labpos.maxdist <- function(pol, gridpoints) 
{
  labw <- 2
  labh <- 2
  makerect <- function(x, y, wd, ht, rectID) 
  {
    CoordX <- c(x-wd/2, x+wd/2, x+wd/2, x-wd/2, x-wd/2)
    CoordY <- c(y-ht/2, y-ht/2, y+ht/2, y+ht/2, y-ht/2)
    Polygons(list(Polygon(cbind(CoordX, CoordY))), rectID)
  }
  getDists <- function(co, pol, pol.l) 
  {
    Srl <- sapply(seq_len(nrow(co)), 
                  function(i) 
                    makerect(co[i, 1], co[i, 2], labw, labh, i))
    rects <- SpatialPolygons(Srl, proj4string = CRS(proj4string(pol)))
    inside <- apply(gContains(pol, rects, byid = TRUE), 1, any)
    di <- numeric(nrow(co))
    if (any(inside)) 
      di[inside] <- gDistance(pol.l, rects[inside], byid = TRUE)
    return(di)
  }
  pol.l <- as(pol, "SpatialLines")
  co <- coordinates(spsample(pol, n = gridpoints, type = "regular"))
  di <- getDists(co, pol, pol.l)
  if (all(di == 0)) 
    stop("Could not fit label inside polygon (with the given number of grid points)")
  origin <- co[which.max(di), ]
  stepsize <- min(apply(co, 2, function(x) diff(sort(unique(x))[1:2])))
  getDistsScaled <- function(co, origin, stepsize, pol, pol.l) 
  {
    co <- matrix(co, ncol = 2)*stepsize+origin
    -getDists(co, pol, pol.l)
  }
  optres <- optim(c(0, 0), getDistsScaled, method = "Nelder-Mead", 
                  control = list(trace = FALSE, reltol = 0.001), origin = origin, 
                  stepsize = stepsize, pol = pol, pol.l = pol.l)
  optval <- optres$par*stepsize+origin
  if (optres$convergence != 0) 
  {
    warning("Numerical optimisation did not converge. Returning result of grid search.")
    optval <- origin
  }
  return(optval)
}
#--------------------------------------------------------------------------------------------
InaccessabityPole <- function(pols, method = c("maxdist", "buffer"), 
                              gridpoints = 60, polypart = c("all", "largest")) 
{
  if (polypart == "largest") 
    pols <- getLargestPolyPart(pols)
  ret <- switch(method, 
                maxdist = labpos.maxdist(pols, gridpoints), 
                buffer  = labpos.buffer(pols))
  return(ret)
}
#--------------------------------------------------------------------------------------------
# Function perform calculating of the pole of inaccessebility coordinates
InscribedRadius <- function(X, SourceCRS, TargetCRC, Id, IterN = 5)
{
  PoleR  <- NA
  # Converting coordinates of the polygon's nodes
  # to the target system of coordinates
  if (class(X)[1] == 'list')
  {
    if (class(X[[1]])[1] == 'Polygon')
      lstPolys <- list(Polygons(srl = X, ID = Id))
    else
      lstPolys <- X
  }
  else
  {
    if (class(X)[1] == 'Polygon')
      lstPolys <- list(Polygons(srl = list(X), ID = Id))
    else
      lstPolys <- list(X)
  }
  SpPolys <- SpatialPolygons(Srl = lstPolys, proj4string = SourceCRS)
  SpPolysProj <- spTransform(x = SpPolys, CRSobj = TargetCRC)
  # The pole of inaccessibility is defined for IterN iterations at 
  # the discretization level increasing twice on each iteration
  for (IterI in 1:IterN)
  {
    ExitCode <- 0
    # Number of grid lines for the search algorithm of a inaccessibility pole
    GridN <- 100*2^(IterI-1)
    PoleXY <- tryCatch(
      # Search of coordinates of a inaccessibility pole
      expr  = InaccessabityPole(pols = SpPolysProj, method = 'maxdist',  
                                gridpoints = GridN, polypart = "all"),
      error = function(e)
      {
        Code <- grep('Could not fit label inside polygon', e$message)
        if (length(Code) > 0)
          # If the error is caused by impossibility to find an inaccessibility pole
          # at this level of discretization
          ExitCode <<- 1
        else
          # If the error is caused by other reason
          ExitCode <<- 2
      }
    )
    if (ExitCode > 0)
      PoleXY <- NULL
    else
      break # If the pole of inaccessibility is found further iterations are cancelled
  }
  if (!is.null(PoleXY))
  {
    # Converting coordinates of a inaccessibility pole 
    # to the SpatialPoints object
    PoleSpP_Proj <- SpatialPoints(coords      = matrix(PoleXY, nrow = 1, ncol = 2), 
                                  proj4string = TargetCRC)
    # points(PoleSpP_Proj, pch = 20, col = 'brown')
    # Sys.sleep(0.1)
    # Converting to the source system of coordinate
    PoleSpPoint <- spTransform(x = PoleSpP_Proj, CRSobj = SourceCRS)
    # Calculation of distance from an inaccessibility pole 
    # to the closest segment of the hole's polygon
    Result <- try(silent = TRUE,
                  expr   = round(dist2Line(p       = PoleSpPoint, 
                                           line    = SpPolys, 
                                           distfun = distGeo)[1]))
    if (class(Result) == 'try-error')
      PoleR <- NA
    else
      PoleR <- Result
  }
  return(PoleR)
}
#-------------------------------------------------------------------------------------
FormatTime <- function(X)
{
  H <- floor(X/3600)
  X <- X-H*3600
  M <- floor(X/60)
  X <- X-M*60
  S <- round(X)
  Str <- sprintf(fmt = '%02i:%02i:%02i', H, M, S)
  return(Str)
}
#-------------------------------------------------------------------------------------
TimeProgress <- function(tic, ExecProgr)
{
  toc <- Sys.time()
  Elapsed <- difftime(toc, tic, tz = 'GMT', units = 'secs')
  Estimated <- as.double(Elapsed*(100/ExecProgr-1))
  Str <- paste(FormatTime(Elapsed), '+', 
               FormatTime(Estimated), '=', 
               FormatTime(Elapsed+Estimated))
  return(Str)
}
#-------------------------------------------------------------------------------------
# Writing data to shape file
WriteShapeData <- function(FileName, LayerName, Polys, CRC, Data, MsgStr)
{
  if (is.null(Polys) | ((length(Polys) == 1) & is.null(Polys[[1]])))
  {
    cat('Файл', FileName, 'не создан — нет данных\n\r')
    return()
  }
  Result <- try(silent = TRUE,
    expr = {
      PO <- 1:length(Polys)
      spPolys <- SpatialPolygons(Srl = Polys, pO = PO, proj4string = CRC)
      remove(PO)
      dfPolys <- SpatialPolygonsDataFrame(Sr       = spPolys, 
                                          data     = Data, 
                                          match.ID = FALSE)
      remove(spPolys)
      rgdal::writeOGR(obj             = dfPolys, 
                      dsn             = dirname(FileName), 
                      layer           = LayerName,
                      driver          = 'ESRI Shapefile', 
                      verbose         = TRUE, 
                      overwrite_layer = TRUE)
    }
  )
  if (class(Result) == 'try-error')
    cat('Ошибка записи в файл', FileName, '\n\r')
  else
    cat(MsgStr, '\n\r')
}
######################################################################################
# MAIN SCRIPT
# Packages obligatory for preliminary installation
library('rgeos')
library('sp')
library('rgdal')
library('geosphere')
# Reading the source data
objShapes1 <- rgdal::readOGR(InFileName, verbose = TRUE)
# CRS object setting the used systems of coordinates and projections
utm24 <- CRS("+proj=utm +zone=24 +ellps=GRS80 +datum=NAD83 +units=m +no_defs +towgs84=0,0,0")
InputCRS <- objShapes1@proj4string
ShN <- length(objShapes1) # Count of shapes
# Calculation of total number of holes in the file
HoleCount <- 0
for (ShI in 1:ShN)
{
  # objPolygons1 - the input multipolygon containing (perhaps) holes
  objPolygons1 <- objShapes1@polygons[[ShI]]@Polygons
  PolN <- length(objPolygons1)
  if (PolN > 1)
    for (PolI in 2:PolN)
      if (objPolygons1[[PolI]]@hole)
        HoleCount <- HoleCount+1
}
cat(HoleCount, 'holes were found\n')
dfShapeMeta <- data.frame(Id       = rep(NA, ShN), 
                          Category = objShapes1@data$cat,
                          Legenda  = objShapes1@data$legenda,
                          Radius   = rep(NA, ShN),
                          Area     = rep(NA, ShN),
                          Perimetr = rep(NA, ShN))
IdFrmLen <- toString(nchar(as.character(ShN)))
IdFrmStr <- paste0('%0', IdFrmLen, 'i')
# lstPolygons2 - the list of output multipolygons 
# containing only multipolygons with large enough holes
lstPolygons2 <- lstBadHoles <- lstBadFragments <- list(NULL)
bhIndex <- bfIndex <- NULL
bhCount <- bfCount <- HoleN <- 0
tic <- Sys.time()
for (ShI in 1:ShN)
{
  objPolygons1 <- objShapes1@polygons[[ShI]]
  lstPolygon1  <- objPolygons1@Polygons
  spPolygons1  <- SpatialPolygons(Srl = list(objPolygons1), proj4string = InputCRS)
  objPolygon1  <- lstPolygon1[[1]]
  lstPolygon2  <- list(objPolygon1) # List of simple polygons including perhaps holes
  strShapeId   <- sprintf(IdFrmStr, ShI)
  dfShapeMeta$Id[ShI] <- strShapeId
  PolN <- length(lstPolygon1)
  if (PolN == 1)
  {
    # If the multipolygon has no simple polygons belonging to it, 
    # then save and pass to the following multipolygon
    lstPolygons2[[ShI]] <- Polygons(list(objPolygon1), ID = paste0(strShapeId, '  1'))
    Progress <- (ShI+HoleN)/(ShN+HoleCount)*100
    cat('\rProgress', sprintf('%2.0f', Progress), '%.', 
        TimeProgress(tic, Progress), ' ')
  }
  else
    for (PolI in 2:PolN)
    {
      # objPolygon1 - the simple polygon as a part of the input multipolygon
      objPolygon1 <- lstPolygon1[[PolI]]
      strPolyId  <- sprintf('%03i',PolI)
      # If the polygon is a hole
      if (objPolygon1@hole)
      {
        PoleR <- NA
        # Calculation of the hole's area (hectares) 
        # for removal of hole with too small, including zero area
        PolyS <- round(areaPolygon(objPolygon1@coords))/1e4
        if (PolyS > MaxHoleArea/10)
        {
          strMultiPolyId <- paste0(strShapeId, strPolyId)
          PoleR <- InscribedRadius(objPolygon1, InputCRS, utm24, strMultiPolyId)
          if (is.na(PoleR))
            # If the pole of inaccessibility hasn't been found
            cat('\rPolygon', toString(ShI), '- Hole', toString(PolI-1), 
                '- It is impossible to find the inaccessibility pole\n')
          # If the hole is big enough it is added to the collected multipolygon
          if ((!is.na(PoleR) & (PoleR > MaxHoleRadius)) | 
               (is.na(PoleR) & (PolyS > MaxHoleArea)))
            lstPolygon2[[length(lstPolygon2)+1]] <- objPolygon1
        }
        if ((is.na(PoleR) & (PolyS > MaxHoleArea)))
        {
          bhCount <- bhCount+1
          lstBadHoles[[bhCount]] <- Polygons(list(objPolygon1), ID = strMultiPolyId)
          bhIndex <- append(bhIndex, ShI)
        }
        HoleN <- HoleN+1
        Progress <- (ShI+HoleN)/(ShN+HoleCount)*100
        cat('\rProgress', sprintf('%2.0f', Progress), '%.', 
            TimeProgress(tic, Progress), ' ')
      } # if (objPolygon@hole)
    } # for (PolI in 1:PolN)
  objPolygons2 <- Polygons(lstPolygon2, ID = strShapeId)
  PoleR <- InscribedRadius(objPolygons1, InputCRS, utm24, strShapeId)
  if (is.na(PoleR))
  {
    # If the pole of inaccessibility hasn't been found
    bfCount <- bfCount+1
    lstBadFragments[[bfCount]] <- objPolygons1
    bfIndex <- append(bfIndex, ShI)
    cat('\r\nPolygon', toString(ShI), 
        '- It is impossible to find the inaccessibility pole')
  }
  PolyS <- round(areaPolygon(spPolygons1))/1e4 # Area in ha
  PolyP <- round(perimeter(spPolygons1))/1e3 # Perimeter in km
  dfShapeMeta$Radius[ShI]   <- PoleR
  dfShapeMeta$Area[ShI]     <- PolyS
  dfShapeMeta$Perimetr[ShI] <- PolyP
  lstPolygons2[[ShI]] <- objPolygons2
} # for (ShI in 1:ShN)
cat('\n')

BadHolesData <- BadFragmentsData <- NULL
if (!(is.null(lstBadHoles) | 
      ((length(lstBadHoles) == 1) & is.null(lstBadHoles[[1]]))))
  BadHolesData <- objShapes1@data[bhIndex, ]
if (!(is.null(lstBadFragments) | 
      ((length(lstBadFragments) == 1) & is.null(lstBadFragments[[1]]))))
  BadFragmentsData <- dfShapeMeta[bfIndex,]
remove(objShapes1)

# Writing output shape file
NameParts <- strsplit(x = basename(OutFileName), split = '.', fixed = TRUE)
WriteShapeData(FileName  = OutFileName,
               LayerName = NameParts[[1]][1],
               Polys     = lstPolygons2,
               CRC       = InputCRS,
               Data      = dfShapeMeta,
               MsgStr    = paste('The output shape file',
                                 OutFileName, 'has been writen'))
remove(lstPolygons2); remove(dfShapeMeta)

# Writing shape file with failure holes
NameParts <- strsplit(x = basename(BadHoleFileName), split = '.', fixed = TRUE)
WriteShapeData(FileName  = BadHoleFileName, 
               LayerName = NameParts[[1]][1], 
               Polys     = lstBadHoles, 
               CRC       = InputCRS, 
               Data      = BadHolesData, 
               MsgStr    = paste('The shape file with bad holes', 
                                 BadHoleFileName, 'has been writen'))
remove(lstBadHoles); remove(BadHolesData)

# Writing shape file with failure fragments
NameParts <- strsplit(x = basename(BadFragmentFileName), split = '.', fixed = TRUE)
WriteShapeData(FileName  = BadFragmentFileName, 
               LayerName = NameParts[[1]][1], 
               Polys     = lstBadFragments, 
               CRC       = InputCRS, 
               Data      = BadFragmentsData, 
               MsgStr    = paste('The shape file with bad framents', 
                                 BadFragmentFileName, 'has been writen'))
# END